/* chm2html.c
    extract html files from chm format,
    base on chmtools written by Matthew T. Russotto

    Copyright (c) 2003 by Samuel Yang <labourvanity@yahoo.com.tw>

    This program is free software;
    You can redistribute it and/or modify it under the terms of
    the GNU Lesser General Public License (version 2) as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this package; if not, contact to me.  */

// version 0.1a

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include "lzx.h"
#include "datatype.h"
#include "chm2html.h"

_S64_ preamble( int *lang_b5){
    char buf[PREAMBLE_LEN];
    _U64_ len;
    if( fread( buf, 1, PREAMBLE_LEN, stdin) != PREAMBLE_LEN)
	CHM_ERR("read preamble error");
    if( strncmp( buf, "ITSF", 4))
	CHM_ERR("not a CHM file");
    if( *(_U32_*)(buf+8) != PREAMBLE_LEN )
	CHM_ERR("illegal preamble size");
    if( *(_U32_*)(buf+20) == LANG_ID_B5)
	*lang_b5= 1;
    else if( *(_U32_*)(buf+20) != LANG_ID)
	CHM_ERR("illegal language ID");
    if( *(_U64_*)(buf+56) != PREAMBLE_LEN )
	CHM_ERR("illegal offset of header 0");
    if( *(_U64_*)(buf+64) != HEADER_0_LEN)
	CHM_ERR("illegal length of header 0");
    if( *(_U64_*)(buf+72) != (PREAMBLE_LEN+HEADER_0_LEN))
	CHM_ERR("illegal offset of header 1");
    len= *(_U64_*)(buf+80);
    if( *(_U64_*)(buf+88) != (len+PREAMBLE_LEN+HEADER_0_LEN))
	CHM_ERR("illegal offset of content 0");
    return len;
}

_S64_ header0( void){
    char buf[HEADER_0_LEN];
    if( fread( buf, 1, HEADER_0_LEN, stdin) != HEADER_0_LEN)
	CHM_ERR("read header 0 error");
    return *(_U64_*)(buf+8);
}

int header1( _U64_ len, char *buf, chunks *chunk){
    _S32_ chunk_num;
    if( fread( buf, 1, len, stdin) != len)
	CHM_ERR("read header 1 error");
    if( strncmp( buf, "ITSP", 4))
	CHM_ERR("illegal header mark");
    if( *(_U32_*)(buf+8) != DIR_LEN)
	CHM_ERR("illegal directory length");
    chunk->size= *(_U32_*)(buf+16);
    chunk->root= *(_U32_*)(buf+28);
    chunk->first= *(_U32_*)(buf+32);
    chunk->last= *(_U32_*)(buf+36);
    chunk_num= *(_U32_*)(buf+44);
    if( *(_U32_*)(buf+48) != LANG_ID)
	CHM_ERR("illegal language ID of header");
    if( (len-DIR_LEN) != (chunk->size*chunk_num))
	CHM_ERR("illegal chunk size");
    if( (chunk->root>=chunk_num) ||
	(chunk->first>=chunk_num) ||
	(chunk->last>=chunk_num))
	    CHM_ERR("illegal chunk number");
    return 0;
}

int chunk_index( char *buf, int num, int size){
    buf += (DIR_LEN+size*num);
    if( strncmp( buf, "PMGI", 4))
	CHM_ERR("illegal chunk index mark");
    return *(_U16_*)(buf+size-2);
}

int chunk_list( char *buf, int num, int size){
    buf += (DIR_LEN+size*num);
    if( strncmp( buf, "PMGL", 4))
	CHM_ERR("illegal chunk list mark");
    return *(_U16_*)(buf+size-2);
}

_U32_ enc_int( char **buf){
    _U32_ i;
    for( i=0; **buf & 0x80; (*buf)++)
	i= (i<<7)+(**buf & 0x7F);
    return (i<<7)+(*(*buf)++);
}

int chunk_entry(  char *buf, int num, int size, entries *entry){
    int i, len;
    char *last;
    buf += (DIR_LEN+size*num);
    last= buf+size-(*(_U16_*)(buf+4));
    len= *(_U16_*)(buf+size-2);
    buf += LIST_OFFSET;
    for( i=0; i<len; i++){
	entry[i].name= buf+1;
	buf += (1+(_U8_)*buf);
	entry[i].section= enc_int(&buf);
	entry[i].offset= enc_int(&buf);
	entry[i].length= enc_int(&buf);
	entry[i].name[entry[i].name[-1]]= 0;
	if( buf >= last)
	    return ++i;
    }
    return i;
}

int read_content( char *buf, _U32_ size){
    if( fread( buf, 1, size, stdin) != size)
	CHM_ERR("read content error");
    if( getchar() != EOF)
	return -1;
    return 0;
}

entries *entry_search( char *name, entries *entry, int entry_num){
    while( entry_num--)
	if( strcmp( entry[entry_num].name, name) == 0)
	    return entry+entry_num;
    return 0;
}

int name_list( char *buf, int size, entries *entry, int entry_num){
    if( (entry=entry_search( NAME_LIST_STR, entry, entry_num)) == 0)
	CHM_ERR("name list not found");
    if( entry->section != 0)
	CHM_ERR("illegal name list section");
    if( entry->length != sizeof(name_list_file))
	CHM_ERR("illegal name list size");
    if( (entry->offset+entry->length) > size )
	CHM_ERR("illegal name list offset");
    buf += entry->offset;
    if( memcmp( buf, name_list_file, sizeof(name_list_file)))
	CHM_ERR("illegal name list file");
    return 0;
}

int content1( char **buf, int size, entries *entry, int entry_num){
    if( (entry=entry_search( CONTENT_1_STR, entry, entry_num)) == 0)
	CHM_ERR("content 1 not found");
    if( entry->section != 0)
	CHM_ERR("illegal content 1 section");
    if( (entry->offset+entry->length) > size)
	CHM_ERR("illegal content 1 size/offset");
    *buf += entry->offset;
    return entry->length;
}

int ctrl_data( char *buf, int size, entries *entry, int entry_num){
    int wlen, window_size;
    if( (entry=entry_search( CTRL_DATA_STR, entry, entry_num)) == 0)
	CHM_ERR("control data not found");
    if( entry->section != 0)
	CHM_ERR("illegal control data section");
    if( entry->length < CTRL_DATA_LEN)
	CHM_ERR("illegal control data size");
    if( (entry->offset+entry->length) > size )
	CHM_ERR("illegal control data offset");
    buf += entry->offset;
    if( strncmp( buf+4, "LZXC", 4))
	CHM_ERR("illegal control data mark");
    wlen= *(_U16_*)(buf+16);
    switch(wlen){
	case 1: window_size = 15; break;
	case 2: window_size = 16; break;
	case 4: window_size = 17; break;
	case 8: window_size = 18; break;
	case 0x10: window_size = 19; break;
	case 0x20: window_size = 20; break;
	case 0x40: window_size = 21; break;
	default: CHM_ERR("illegal window size");
    }
    return window_size;
}

_S64_ span_info( char *buf, int size, entries *entry, int entry_num){
    if( (entry=entry_search( SPAN_INFO_STR, entry, entry_num)) == 0)
	CHM_ERR("span information not found");
    if( entry->section != 0)
	CHM_ERR("illegal span information section");
    if( entry->length != 8)
	CHM_ERR("illegal span information size");
    if( (entry->offset+entry->length) > size)
	CHM_ERR("illegal span information offset");
    buf += entry->offset;
    return *(_S64_*)buf;
}

void utf8_to_bg5( char *buf, char *root, char *name){
    _U32_ unicode;
    int bit, extras;
    while( *root)
	*buf++ = *root++;
    while( *name)
	if( (*name & 0x80) == 0)
	    *buf++ = *name++;
	else if( (*name & 0xC0) == 0x80){
	    *buf++ = '?';
	    name++;
	}else{
	    for( extras=1, bit=0x20; (*name & bit) != 0; bit >>= 1)
		extras++;
	    if( extras > 5){
		*buf++ = '?';
		name++;
	    }
	    for( unicode=(*name++ & --bit); extras; extras--)
		if( (*name & 0x80) == 0){
		    *buf++ = '?';
		    break;
		}else
		    unicode= (unicode<<6)+(*name++ & 0x3F);
	    if( extras == 0){
		unicode= b5_out[unicode];
		//sprintf( buf, "%%%02X%%%02X",unicode>>8,unicode&255);
		sprintf( buf, "%c%c",unicode>>8,unicode&255);
		buf += strlen(buf);
	    }
	}
    *buf=0;
}

int extract( int lang_b5, char *root, char *buf, int size, entries *entry, int entry_num){
    int i,err;
    char path[1024];
    FILE *f;
    for( i=err=0; i<entry_num; i++){
	if( lang_b5)
	    utf8_to_bg5( path, root, entry[i].name);
	else
	    sprintf( path, "%s%s", root, entry[i].name);
	if( entry[i].name[0] != '/')
	    continue;
	else if( entry[i].section == 0){
	    if( mkdir( path, 0x1ED) && (errno != EEXIST)){
		fprintf( stderr, "create directory error: '%s'\n", path);
		err++;
	    }
	}else if( (entry[i].offset+entry[i].length) > size){
	    fprintf( stderr, "illegal file: '%s'\n", entry[i].name);
	    err++;
	}else{
	    if( (f=fopen( path, "wb")) == 0){
		fprintf( stderr, "open file error: '%s'\n", path);
		err++;
		continue;
	    }
	    if( fwrite( buf+entry[i].offset,1,entry[i].length,f) != entry[i].length){
		fprintf( stderr, "write file error: '%s'\n", path);
		err++;
	    }
	    fclose(f);
	}
    }
    return err;
}

int main( int argc, char *argv[]){
    _S64_ header_len;
    _S64_ file_size;
    _S64_ content_size;
    _S64_ span_size;
    _S64_ cmpress_size;
    char *header, *content, *cmpress, *span;
    chunks chunk;
    int entry_num;
    entries *entry;
    int i,j,k, err, lang_b5;
    int window_size;
    char *root;
    if( (header_len=preamble( &lang_b5)) < 0)
	return 1;
    if( (file_size=header0()) < 0)
	return 1;
    if( (header=malloc( header_len)) == 0){
	fputs( "insufficient memory\n", stderr);
	return 1;
    }
    if( header1( header_len, header, &chunk) < 0)
	return 1;
    if( chunk.root >= 0){
	if( (j=chunk_index( header, chunk.root, chunk.size)) < 0)
	    return 1;
	if( j != (chunk.last-chunk.first+1)){
	    fputs( "illegal chunk index\n", stderr);
	    return 1;
	}
    }
    for( entry_num=0, i=chunk.first; i<=chunk.last; i++)
	if( (j=chunk_list( header, i, chunk.size)) < 0)
	    return 1;
	else
	    entry_num += j;
    if( (entry=malloc( sizeof(entries)*entry_num)) == 0){
	fputs( "insufficient memory\n", stderr);
	return 1;
    }
    for( k=0, i=chunk.first; i<=chunk.last; i++)
	if( (j=chunk_entry( header, i, chunk.size, entry+k)) < 0)
	    return 1;
	else
	    k += j;
    if( k != entry_num ){
	fputs( "illegal chunk list\n", stderr);
	return 1;
    }
    content_size= file_size-header_len-(PREAMBLE_LEN+HEADER_0_LEN);
    if( (content=malloc( content_size)) == 0){
	fputs( "insufficient memory\n", stderr);
	return 1;
    }
    if( read_content( content, content_size) < 0)
	return 1;
    if( name_list( cmpress=content, content_size, entry, entry_num) < 0)
	return 1;
    if( (cmpress_size=content1( &cmpress, content_size, entry, entry_num)) < 0)
	return 1;
    if( (window_size=ctrl_data( content, content_size, entry, entry_num)) < 0)
	return 1;
    if( (span_size=span_info( content, content_size, entry, entry_num)) < 0)
	return 1;
    if( (span=malloc( span_size)) == 0){
	fputs( "insufficient memory\n", stderr);
	return 1;
    }
    LZXinit( window_size);
    if( err= LZXdecompress( cmpress, span, cmpress_size, span_size)){
	fprintf( stderr, "error code: %d\n", err);
	return 1;
    }
    root= (argc>1)?argv[1]:".";
    return extract( lang_b5, root, span, span_size, entry, entry_num)?2:0;
}
